(* Calendar module. The calendar displays various items; each item has a title,
   a date on which it should be displayed, and optionally links to some page for
   that item (for example, an assignment would link to the page for that
   assignment. *)

structure BS = Bootstrap3
structure DT = Datetime

style calendar
style date_cell
style month_links
style prev_month
style current_month
style next_month
style date_number

type item = {
     Title : string,
     Datetime : DT.t,
     Link : url
}

fun onDay (date : DT.t) (datetime : DT.t) =
    date.Year = datetime.Year &&
    date.Month = datetime.Month &&
    date.Day = datetime.Day

(* Construct a range of dates by adding 1 to start until it's >= finish (i.e.,
   inclusive of the lower bound, exclusive of the upper bound. *)

fun dateRange start finish =
    if start >= finish then []
    else start :: dateRange (DT.addDays 1 start) finish

fun render (item : item) : xbody =
    <xml><li><a href={item.Link}>{[item.Title]}</a></li></xml>

fun splitEvery [a] (n : int) (xs : list a) : list (list a) =
    case List.splitAt n xs of
        (_, []) => xs :: []
      | (part, rest) => part :: splitEvery n rest

val dayOfWeekHeader = <xml><thead><tr>
  {List.mapX (fn dayOfWeek => <xml><th>{[dayOfWeek]}</th></xml>)
   (DT.Sunday :: DT.Monday :: DT.Tuesday :: DT.Wednesday :: DT.Thursday ::
              DT.Friday :: DT.Saturday :: [])}
  </tr></thead></xml>

functor Make(M : sig
                 val getItems : transaction (list item)
                 val template : Template.t -> transaction page
             end) = struct

    fun showItems (year : int) (month : DT.month) : transaction page =
        let
            val firstOfMonth = DT.normalize {
                               Year = year,
                               Month = month,
                               Day = 1,
                               Hour = 0,
                               Minute = 0,
                               Second = 0
                               }
            val rangeStart = DT.addDays
                                 (-(DT.dayOfWeekToInt
                                        (DT.dayOfWeek firstOfMonth)))
                                 firstOfMonth
            val rangeEnd = DT.addDays 35 rangeStart
            val weeks = splitEvery 7 (dateRange rangeStart rangeEnd)
            val prevLink = let
                val prevMonthYear =
                    if month = DT.January then year - 1 else year
                val prevMonth = if month = DT.January then DT.December else
                                DT.intToMonth (DT.monthToInt month - 1)
            in
               <xml><a class={prev_month}
               link={showItems prevMonthYear prevMonth}>
               &larr; {[prevMonth]} {[prevMonthYear]}
               </a></xml>
            end
            val nextLink = let
                val nextMonthYear =
                    if month = DT.December then year + 1 else year
                val nextMonth = if month = DT.December then DT.January else
                                DT.intToMonth (DT.monthToInt month + 1)
            in
               <xml><a class={next_month}
               link={showItems nextMonthYear nextMonth}>
               {[nextMonth]} {[nextMonthYear]} &rarr;
               </a></xml>
            end
            fun renderDate items date = <xml>
              <td class={date_cell}>
                <div>
                  <p class="date-number">{[DT.format "%e" date]}</p>
                  <ul>
                    {List.mapX render (List.filter
                                           (fn item => onDay date item.Datetime)
                                           items)}
                  </ul>
                </div>
              </td>
            </xml>
            fun renderWeek items week = <xml>
              <tr>
                {List.mapX (renderDate items) week}
              </tr>
            </xml>
        in
            items <- M.getItems;
            M.template {
            Inner = <xml>
              <span style="display: inline-block;">
                <div class={month_links}>
                  {prevLink}
                  <a class={current_month} link={showItems year month}>
                    {[month]} {[year]}
                  </a>
                  {nextLink}
                </div>
                <table class="calendar">
                  {dayOfWeekHeader}
                  {List.mapX (renderWeek items) weeks}
                </table>
              </span>
            </xml>,
            Crumbs = {Url = None, Title = "Calendar"} :: []}
        end

    fun currentCalendar () =
        dt <- DT.now;
        showItems dt.Year dt.Month
end
