(* A Gradable represents something that students will be graded on. It can
   represent both assignments, for which the teacher and the student both need
   to upload files, and class participation, for which neither the teacher nor
   the student needs to upload anything. *)

(* Necessary because you can't do [REFERENCES Authentication.userTbl]. *)
open Authentication
open Bootstrap3
open Template
open Util

structure Auth = Authentication
structure BS = Bootstrap3
structure L = List
structure V = Variant

con id = Material.id
con component = [Id = int, Gradable = serialized id, Index = int,
                 Description = string, MaxGrade = int]
con gradeCols = [Component = int, Username = string, Grade = int]
con submissionCols = [Gradable = serialized id, Username = string,
                      Submitted = time]
con gradableCols = [DueDate = time, Weight = int]
con routes = Material.routes ++ [Submit = id, Components = id, Grades = id,
                                 ViewSubmission = id * string]

fun validateWeight weight = if weight < 0
                            then Some "Weight must be non-negative"
                            else None

functor Make(M : sig
    con submissionData :: {Type}
    constraint submissionData ~ submissionCols
    val submissionFl : folder submissionData
    val submissionInj : $(map sql_injectable submissionData)
    val validateSubmission : $submissionData -> transaction (result {})

    con gradableData :: {Type}
    val gradableFl : folder gradableData
    constraint Material.cols ~ gradableData
    constraint gradableCols ~ gradableData
    val validateGradable : $gradableData -> transaction (result {})
  end) = struct
    structure Mat = Material.Make(struct
        con cols = M.gradableData ++ gradableCols
        val fl = @Folder.concat ! M.gradableFl _
        val validate = fn x =>
                          v <- M.validateGradable (project [M.gradableData] x);
                          let val v' = case validateWeight x.Weight of
                                           Some err => Error err
                                         | None => Success {}
                          in return (v; v')
                          end
      end)

    con t' = Mat.t'
    con t = Mat.t

    table gradeTbl : gradeCols
        PRIMARY KEY (Component, Username)

    sequence componentIds

    table componentTbl : component
        PRIMARY KEY (Id)

    con submission = submissionCols ++ M.submissionData
    table submissionTbl : submission
        PRIMARY KEY (Gradable, Username)
        CONSTRAINT Username
          FOREIGN KEY Username REFERENCES userTbl (Username)
          ON DELETE CASCADE

    val create = Mat.create
    val lookup = Mat.lookup
    val update = Mat.update
    val delete = Mat.delete
    val list = Mat.list

    con userSubData = [UserTbl = user, SubmissionTbl = map option submission]


    (* Get all grades for a given materialId; this will have one row for each
       (user, component) pair. *)
    con gradeData = {Username : string, Grade : option int,
                     Component : int, MaxGrade : int}
    fun gradesFor (id : id) : transaction (list gradeData) =
        results <- queryL (SELECT userTbl.Username, componentTbl.Id,
                             gradeTbl.Grade, componentTbl.MaxGrade
                           FROM (userTbl CROSS JOIN componentTbl)
                             LEFT OUTER JOIN gradeTbl
                             ON (gradeTbl.Username = userTbl.Username
                               AND gradeTbl.Component = componentTbl.Id)
                          ORDER BY componentTbl.Index DESC);
        return (List.mp (fn row => { Username = row.UserTbl.Username,
                                     Grade = row.GradeTbl.Grade,
                                     Component = row.ComponentTbl.Id,
                                     MaxGrade = row.ComponentTbl.MaxGrade })
               results)

    val tbl = Mat.tbl
    fun gradeSummary user =
        grades <-
        List.mapQuery (SELECT componentTbl.Gradable,
                         SUM(gradeTbl.Grade), SUM(componentTbl.MaxGrade)
                       FROM componentTbl LEFT OUTER JOIN gradeTbl
                         ON (gradeTbl.Username = {[user.Username]}
                           AND gradeTbl.Component = componentTbl.Id)
                       GROUP BY componentTbl.Gradable)
        (fn row => {Gradable = deserialize row.ComponentTbl.Gradable,
                    Grade = row.1,
                    MaxGrade = row.2});
        gradables <- list;
        (* Sort of an ugly hack; for each gradable, we want the sum of all the
           student's grades for that gradable, as well as its title, which
           requires something like

           SELECT ... FROM (SELECT ... FROM componentTbl LEFT OUTER JOIN
           gradeTbl ON ... GROUP BY ...

           which you can't do in Ur/Web. So here we rely on the fact that
           Act.list and the above query return results in the same order
           (ascending by gradable ID). *)
        return (ListPair.mp
                    (fn grade gradable => grade ++ {Title = gradable.Title,
                                                    Weight = gradable.Weight})
                grades gradables)

    (* Get a student's grades for a given material. *)
    fun gradeFor id username =
        List.mapQuery (SELECT componentTbl.Description,
                         componentTbl.MaxGrade, componentTbl.Id,
                         gradeTbl.Grade
                       FROM componentTbl LEFT OUTER JOIN gradeTbl
                         ON componentTbl.Id = gradeTbl.Component
                         AND gradeTbl.Username = {[username]}
                       WHERE componentTbl.Gradable = {[serialize id]}
                       (* This is ASC because mapQuery returns results in query
                       order. *)
                       ORDER BY componentTbl.Index ASC)
                 (fn row => { Description = row.ComponentTbl.Description,
                              Component = row.ComponentTbl.Id,
                              Grade = row.GradeTbl.Grade,
                              MaxGrade = row.ComponentTbl.MaxGrade  })

    fun gradesHierarchy id
        : transaction (list {GradeRow : list {Grade : option int,
                                              Component : int}}) =
        students <- getStudents;
        List.mapM (fn {Username = username, ...} =>
            grades <- gradeFor id username;
            return {GradeRow = List.mp (
                    fn grade => {Grade = grade.Grade,
                                 Component = grade.Component}) grades})
        students


    fun componentsFor (id : id) =
        queryL1 (SELECT * FROM componentTbl
                          WHERE componentTbl.Gradable = {[serialize id]}
                            ORDER BY componentTbl.Index DESC)

    (* Get a user's submission for an material. *)
    fun submissionFor (id : id) (username : string)
        : transaction (option $submission) =
        oneOrNoRows1 (SELECT * FROM submissionTbl
                               WHERE (submissionTbl.Gradable = {[serialize id]}
                               AND submissionTbl.Username = {[username]}))

    (* Update the data for a user's submission. *)
    fun updateSubmission (id : id)
                         (user : $Authentication.user)
                         (submission : $M.submissionData) : transaction unit =
        n <- now;
        let
            val exp = @Sql.sqexps (@Folder.concat ! M.submissionFl _)
                       (M.submissionInj ++ _)
                       (submission ++ { Gradable = serialize id,
                                        Username = user.Username,
                                        Submitted = n })
            val sel = Sql.selector [#T] {Username = user.Username,
                                          Gradable = serialize id}
        in
            dml (Basis.delete submissionTbl sel);
            dml (Basis.insert submissionTbl exp)
        end

    functor Gui(R : sig
        val name : string
        con auxRoutes :: {Type}
        val auxRoutesFl : folder auxRoutes
        constraint routes ~ auxRoutes
        val auxHandlers : Linkage.handlers (routes ++ auxRoutes) auxRoutes
        val viewGradable : Linkage.router (routes ++ auxRoutes) ->
                           option id -> $M.gradableData -> transaction xbody
        val viewSubmission : Linkage.router (routes ++ auxRoutes) ->
                             option id ->
                             $(M.submissionData ++ submissionCols) ->
                             transaction xbody
        con formState :: Type
        val formEls : Gui.formElement (routes ++ auxRoutes)
                                      (option id) M.gradableData formState body
        type submissionState
        val submissionEl : Gui.formElement (routes ++ auxRoutes)
                           (option $submissionCols)
                           M.submissionData submissionState body
      end) = struct
        fun listCrumb (r : Linkage.router (routes ++ R.auxRoutes)) =
                       {Url = Some (r (make [#ViewAll] ())), Title = R.name}

        (* If the submission data is an empty record, students don't have
           anything to submit so nothing should be displayed. *)
        val hasSubmission =
            @fold [fn _ => bool] (fn [nm ::_] [v ::_] [r ::_] [[nm] ~ r]
                                              _ => True)
            False M.submissionFl

        fun viewSubmission r (id, username) =
            requireTeacherOr username;
            submission <- submissionFor id username;
            case submission of
                Some submission' =>
                rendered <- R.viewSubmission r (Some id) submission';
                template {
                Inner = rendered,
                Crumbs = listCrumb r ::
                         {Url = Some (r (make [#ViewOne] id)),
                          Title = "View"} :: []
                }
              | None => error <xml>404</xml>


        fun gradesFolder g acc = {
            Xml = join acc.Xml <xml><tr>
              <td>{[g.Description]}</td>
              <td>{[case g.Grade of
                        Some g' => show g'
                      | None => ""]}</td>
              <td>{[g.MaxGrade]}</td>
              <td>{[case g.Grade of
                        Some g' => show (round
                                   (100.0 * float g' / float g.MaxGrade))
                      | None => ""]}</td>
            </tr>
            </xml>,
            Grade = Option.get 0 g.Grade + acc.Grade,
            MaxGrade = (if Option.isSome g.Grade then g.MaxGrade else 0) +
                       acc.MaxGrade
        }

        (* Create a table of the current user's grades for a gradable if they're
           a student. *)
        fun showGrades id = bind (Auth.withUser (fn user =>
              grade <- gradeFor id user.Username;
              let
                  val gradeData = List.foldl gradesFolder
                                  {Xml = <xml/>, Grade = 0, MaxGrade = 0}
                                  grade
              in
              return <xml><div class="row"><div class="col-sm-6">
                <table class="bs3-table">
                  <thead>
                    <tr>
                      <th>Component</th><th>Grade</th><th>Maximum</th>
                      <th>Percent</th>
                    </tr>
                  </thead>
                  <tbody>
                    {gradeData.Xml}
                    <tr>
                      <td>Total</td>
                      <td>{[gradeData.Grade]}</td>
                      <td>{[gradeData.MaxGrade]}</td>
                      <td>{[if gradeData.MaxGrade = 0 then "" else show (round (
                            100.0 * float gradeData.Grade /
                                    float gradeData.MaxGrade))]}</td>
                    </tr>
                  </tbody>
                </table>
              </div></div></xml> end)) ifStudent

        fun submit r id =
            Auth.withUser (fn user => Mat.withMaterial id (fn gradable =>
            gradeXml <- showGrades id;
            submission <- submissionFor id user.Username;
            state <- R.submissionEl.Init r
                       (Option.mp (fn x => x --- M.submissionData) submission)
                       (Option.mp (fn x => x --- submissionCols) submission);
            saveButton <- Widget.formElSaveButton R.submissionEl state
                   (fn v => rpc (updateSubmission id user v));
            template {
            Inner = <xml>
              {R.submissionEl.Edit state}
              {gradeXml}
              {saveButton}
            </xml>,
            Crumbs = listCrumb r ::
                     {Url = Some (r (make [#ViewOne] id)),
                      Title = gradable.Title} ::
                     {Url = None, Title = "Submit"} :: []}))

        fun submitLink r id =
            if hasSubmission then
                Auth.ifStudent (case id of
                    Some id' => <xml>
                      <a class="btn btn-primary" href={r (make [#Submit] id')}>
                        Submit
                    </a></xml>
                  | None => <xml/>)
            else return <xml/>

        (* Turn a list of (username, grade source) pairs either into None, which
           indicates that one of the grades is invalid, or Some list of
           (username, grade) pairs. An empty grade indicates that the assignment
           hasn't been graded. *)
        fun extract (sources : list (string * source string))
            : signal (option (list (string * option int))) =
            List.foldlM (fn fragment acc =>
            case acc of
                Some acc' =>
                grade <- signal fragment.2;
                return (case grade of
                    "" => Some ((fragment.1, None) :: acc')
                   | _ => case read grade of
                         Some grade' => Some ((fragment.1, Some grade') :: acc')
                       | None => None)
              | None => return None) (Some []) sources


        fun updateGrades newGrades = List.app (fn gradeData =>
            List.app (fn grade =>
                dml (DELETE FROM gradeTbl
                              WHERE Username = {[gradeData.Username]}
                                AND Component = {[grade.Component]});
                case grade.Grade of
                    Some grade' => dml (
                              INSERT INTO gradeTbl (Username, Component, Grade)
                              VALUES ({[gradeData.Username]},
                                      {[grade.Component]},
                                      {[grade']}))
                  | None => return ()) gradeData.GradeRow) newGrades

        (* Parse a grade and bounds-check it. The empty string is valid, and
           corresponds to no grade. *)
        fun parseGrade maxGrade grade = if grade = ""
                                        then Util.Success None else
            case read grade : option int of
                Some grade' => if grade' >= 0 && grade' <= maxGrade
                               then Util.Success (Some grade')
                               else Util.Error "Out of bounds"
              | None => Util.Error "Not a valid integer"


        (* An element corresponding to a student's grades for an individual
           assignment. The metadata is the username and a list of max grades;
           the data is a list of (component id, grade) values. The lists must be
           in the same order and match up. *)
        val gradeEl
            : Gui.formElement _ _ [GradeRow = list {Component : int,
                                                    Grade : option int}] _ _ = {
            Init = fn r meta grades =>
                      gradeInfos <- listPairMapM (
                                    fn maxGrade gr =>
                                       src <- source (case gr.Grade of
                                                          gr' => show gr'
                                                        | None => "");
                                       return {MaxGrade = maxGrade,
                                               Source = src,
                                               Component = gr.Component})
                                                 meta.MaxGrades
                                                 (getOptField [#GradeRow] []
                                                              grades);
                      return {SubmissionLink = <xml>
                      <a href={r (make [#ViewSubmission] (meta.Id,
                                                          meta.Username))}>
                        {[meta.Username]}
                      </a></xml>, Grades = gradeInfos},
            Signals = fn {Grades = grades, ...} =>
                results <- List.mapM (
                           fn grade =>
                              gr <- signal grade.Source;
                              return (gr' <- parseGrade grade.MaxGrade gr;
                                      return {Grade = gr',
                                              Component = grade.Component}))
                                     grades;
                return (results' <- Gui.perform results;
                        return {GradeRow = results'}),
            Edit = fn state => <xml>
              <td>{state.SubmissionLink}</td>
              {List.mapX (
               fn grade => <xml>
                 <td>
                   <ctextbox class="form-control" source={grade.Source} size=2/>
                 </td>
               </xml>) state.Grades}
            </xml>,
            OnSave = fn _ => return {},
            (* TODO: This should do actual validation; unfortunately, that's not
               possible because we don't have access to the max grades. So maybe
               meta-info should be passed to Validate? *)
            Validate = fn v => return (Success {})
        }

        val gradesEl = Gui.tableEl [#Grades] False gradeEl

        fun editGrades r id =
            Auth.requireTeacher;
            Mat.withMaterial id (fn gradable =>
            hier <- gradesHierarchy id;
            students <- Auth.getStudents;
            components <- componentsFor id;
            let
                val headers = <xml/> ::
                    (List.mp (fn component =>
                    <xml>
                      {[component.Description]} (out of {[component.MaxGrade]})
                    </xml>) components)
                val maxGrades = List.mp (fn component => component.MaxGrade)
                                components
                val meta = List.mp (fn user =>
                                       {Username = user.Username,
                                        MaxGrades = maxGrades,
                                        Id = id}) students
            in
                state <- gradesEl.Init r
                            {Headers = headers,
                             Default = {Username = "???", Id = id,
                                        MaxGrades = []},
                             Meta = meta}
                            (Some {Grades = hier});
                saveButton <- Widget.formElSaveButton gradesEl state
                        (fn v => rpc (updateGrades (ListPair.mp (
                            fn user grade => {Username = user.Username,
                                              GradeRow = grade.GradeRow})
                                                                students
                                                                v.Grades)));
                template {
                Inner = <xml>
                  <div>{gradesEl.Edit state}</div>
                  <div>{saveButton}</div>
                </xml>,
                Crumbs = listCrumb r ::
                         {Url = Some (r (make [#Grades] id)),
                          Title = gradable.Title} ::
                         {Url = None, Title = "Edit grades"} :: []
                }
             end)

        fun gradeLink r id = Auth.ifTeacher
            (case id of
                Some id' => <xml>
                  <a class="btn btn-default" href={r (make [#Grades] id')}>
                    Grade
                </a></xml>
              | None => <xml/>)

        val componentEl = Gui.recordRow {
            Description = {
            Default = "New component",
            Stringify = fn x => x,
            Parse = Util.Success,
            Validate = fn x =>
                if x = "" then Error "Description must not be empty"
                else Success {}},
            MaxGrade = {
            Default = 100,
            Stringify = show,
            Parse = fn x => case read x of
                                Some x' => Success x'
                              | _ => Error "Must be an integer",
            Validate = fn x => if x < 0
                               then Error "Max grade must be positive"
                               else Success {}
            }
            }
        val componentsEl =
            Gui.tableEl [#Components] True componentEl

        fun setComponents gradable components =
            Auth.requireTeacher;
            dml (DELETE FROM componentTbl
                        WHERE Gradable = {[serialize gradable]});
            let
                fun inserter index xs = case xs of
                    x :: xs' =>
                    id <- nextval componentIds;
                    dml (INSERT INTO componentTbl (Id, Gradable, Index,
                             Description, MaxGrade)
                         VALUES ({[id]}, {[serialize gradable]}, {[index]},
                             {[x.Description]}, {[x.MaxGrade]}));
                    inserter (index + 1) xs'
                  | [] => return {}
            in
                inserter 0 components
            end


        fun editComponents r id =
            Auth.requireTeacher;
            Mat.withMaterial id (fn gradable =>
            components <- componentsFor id;
            state <- componentsEl.Init r
                    {Headers = cdata "Description" :: cdata "Grade" :: [],
                     Meta = List.mp (fn _ => {}) components,
                     Default = {}}
                    (Some {Components = List.mp (
                           fn comp => {Description = comp.Description,
                                       MaxGrade = comp.MaxGrade})
                                                components});
            saveButton <- Widget.formElSaveButton componentsEl state
                         (fn v' => rpc (setComponents id v'.Components));
            template {
            Inner = <xml>
              <div>{componentsEl.Edit state}</div>
              <div>{saveButton}</div>
            </xml>,
            Crumbs = listCrumb r ::
                     {Url = Some (r (make [#ViewOne] id)),
                      Title = gradable.Title} ::
                     {Url = None, Title = "Edit components"} :: []
            })

        fun componentsLink r id = Auth.ifTeacher
            (case id of
                Some id' => <xml>
                  <a class="btn btn-default" href={r (make [#Components] id')}>
                    Edit components
                </a></xml>
              | None => <xml/>)

        val weightEl = Gui.intEl [#Weight] "Weight" 0 validateWeight

        val dueDateEl = Gui.dateEl [#DueDate] "Due date"
                                   {Year = 2014, Month = Datetime.January,
                                    Day = 1, Hour = 0, Minute = 0, Second = 0}
                                   (fn _ => None)

        structure MatGui = Mat.Gui(struct
            val name = R.name
            val viewMaterial = fn r id cols =>
                gradableXml <- R.viewGradable r id (cols --- gradableCols);
                gradeXml <- (case id of
                                Some id' => showGrades id'
                              | None => return <xml/>);
                submitLink <- submitLink r id;
                gradeLink <- gradeLink r id;
                componentsLink <- componentsLink r id;
                return <xml>
                  {gradableXml}
                  {gradeXml}
                  <div>
                    {submitLink}{gradeLink}{componentsLink}
                  </div>
                </xml>
            val auxHandlers = {Submit = submit, Components = editComponents,
                               Grades = editGrades,
                               ViewSubmission = viewSubmission} ++
                              R.auxHandlers
            val fl = @Folder.concat ! R.auxRoutesFl _
            val formEl = Gui.concatForm (Gui.concatForm R.formEls dueDateEl)
                                        weightEl
         end)
        val routed = MatGui.routed
    end
end
