open Bootstrap3
open Util

structure L = Linkage

fun recordM [m ::: (Type -> Type)] (inst : monad m) [nm :: Name] [a ::: Type]
    (v : m a) : m {nm : a} = v' <- v; return { nm = v' }

fun copy [t ::: Type] (src : source t) (dst : source t) =
    v <- get src; set dst v

fun optToResult (x : option string) : result {} =
    case x of
        Some err => Error err
      | None => Success ()

fun resultToOpt x =
    case x of
        Success _ => None
      | Error err => Some err

con formElement = fn (routes :: {Type}) (meta :: Type) (fs :: {Type})
                                        (state :: Type) (ctx :: {Unit}) => {
                     Init : L.router routes -> meta -> option $fs ->
                            transaction state,
                     Signals : state -> signal (result $fs),
                     Edit : state -> xml ctx [] [],
                     OnSave : state -> transaction unit,
                     Validate : $fs -> transaction (result {})
                     }

fun concatForm [routes ::: {Type}] [meta ::: Type] [fs1 ::: {Type}]
               [state1 ::: Type] [fs2 ::: {Type}] [state2 ::: Type] [fs1 ~ fs2]
               [ctx ::: {Unit}]
               (form1 : formElement routes meta fs1 state1 ctx)
               (form2 : formElement routes meta fs2 state2 ctx) = {
    Init = fn r meta x =>
              case x of
                  Some x' =>
                  init1 <- form1.Init r meta (Some (x' --- fs2));
                  init2 <- form2.Init r meta (Some (x' --- fs1));
                  return (init1, init2)
                | None =>
                  init1 <- form1.Init r meta None;
                  init2 <- form2.Init r meta None;
                  return (init1, init2),
    Signals = fn state =>
                 signals1 <- form1.Signals state.1;
                 signals2 <- form2.Signals state.2;
                 return (s1 <- signals1;
                         s2 <- signals2;
                         return (s1 ++ s2)),
    Edit = fn state => join (form1.Edit state.1) (form2.Edit state.2),
    OnSave = fn state => form1.OnSave state.1; form2.OnSave state.2,
    Validate = fn data =>
      v1 <- form1.Validate (data --- fs2);
      v2 <- form2.Validate (data --- fs1);
      return (v1; v2; return ())
}

fun dataSignal [t ::: Type] (inputSrc : source string)
               (parse : string -> result t) (validate : t -> option string)
    : signal (result t) =
    src' <- signal inputSrc;
    case parse src' of
        Error err => return (Error "Could not parse data")
      | Success v => case validate v of
                         None => return (Success v)
                       | Some err => return (Error err)

fun errorMessage [t ::: Type] (dataSignal : signal (result t))
    : signal xbody =
    data <- dataSignal;
    case data of
        Error err => return <xml>{[err]}</xml>
      | Success _ => return <xml/>

(* Given a signal representing a result, generate a signal representing the
   corresponding Bootstrap CSS class for success/error. *)
fun colorize [t ::: Type] (dataSrc : signal (result t)) =
    data <- dataSrc;
    return (case data of
                Success _ => CLASS "has-success"
              | Error _ => CLASS "has-error")

(* A textbox element. *)
fun textboxEl [routes ::: {Type}] [a ::: Type] [nm :: Name] (lbl : string)
              (def : string) (validate : string -> option string) :
    formElement routes a [nm = string] (source string) body =
    { Init = fn _ _ str => source (getOptField [nm] def str),
      Signals = fn src =>
                   str <- dataSignal src Success validate;
                   return (recordM [nm] str),
      Edit = fn src =>
        <xml>
          <div dynClass={colorize (dataSignal src Success validate)}>
            <label class="control-label">{[lbl]}</label>
            <ctextbox class="form-control" source={src}/>
            <p class="help-block">
              <dyn signal={errorMessage (dataSignal src Success validate)}/>
            </p>
          </div>
        </xml>,
      OnSave = fn _ => return {},
      Validate = fn x => return (optToResult (validate x.nm))
}

(* A textarea element. *)
fun textareaEl [routes ::: {Type}] [a ::: Type] [nm :: Name] (lbl : string)
               (def : string) (validate : string -> option string) :
    formElement routes a [nm = string] (source string) body =
    { Init = fn _ _ str => source (getOptField [nm] def str),
      Signals = fn src =>
                   str <- dataSignal src Success validate;
                   return (recordM [nm] str),
      Edit = fn src =>
        <xml>
          <div dynClass={colorize (dataSignal src Success validate)}>
            <label class="control-label">{[lbl]}</label>
            <ctextarea class="form-control" source={src}/>
            <p class="help-block">
              <dyn signal={errorMessage (dataSignal src Success validate)}/>
            </p>
          </div>
        </xml>,
      OnSave = fn _ => return {},
      Validate = fn x => return (optToResult (validate x.nm))
}

fun readInt str = case read str of
                      Some n => Success n
                    | None => Error "Not an integer"

(* An element representing an integer. *)
fun intEl [routes ::: {Type}] [a ::: Type] [nm :: Name] (lbl : string)
          (def : int) (validate : int -> option string) :
    formElement routes a [nm = int] (source string) body =
    { Init = fn _ _ int => source (show (getOptField [nm] def int)),
      Signals = fn src =>
                   int <- dataSignal src readInt validate;
                   return (recordM [nm] int),
      Edit = fn src =>
        <xml>
          <div dynClass={colorize (dataSignal src readInt validate)}>
            <label class="control-label">{[lbl]}</label>
            <ctextbox class="form-control" source={src}/>
            <p class="help-block">
              <dyn signal={errorMessage (dataSignal src readInt validate)}/>
            </p>
          </div>
        </xml>,
      OnSave = fn _ => return {},
      Validate = fn x => return (optToResult (validate x.nm))
}

fun dateEl [routes ::: {Type}] [a ::: Type] [nm :: Name] (lbl : string)
           (def : Datetime.t) (validate : Datetime.t -> option string) :
    formElement routes a [nm = time] _ body = let
        fun dateSignal picker =
            dt <- DatetimePicker.value picker;
            return (case dt of
                        None => Error "Must enter valid date"
                      | Some dt' => Success (Datetime.toTime dt'))
    in
    { Init = fn r _ t => DatetimePicker.create (
                         case t of
                             Some t' => Some (Datetime.fromTime t'.nm)
                           | None => None),
      Signals = fn picker =>
                   date <- dateSignal picker;
                   return (recordM [nm] date),
      Edit = fn picker =>
        <xml>
          <div dynClass={Monad.mp (classes (CLASS "form-group"))
                                  (colorize (dateSignal picker))}>
            <label class="control-label">{[lbl]}</label>
            {DatetimePicker.render picker}
            <p class="help-block">
              <dyn signal={errorMessage (dateSignal picker)}/>
            </p>
          </div>
        </xml>,
      OnSave = fn _ => return (),
      Validate = fn t => return (optToResult (validate (Datetime.fromTime t.nm)))
    }
    end

fun xmlEl [routes ::: {Type}] [a ::: Type]
          (genXml : L.router routes -> a -> transaction xbody)
    : formElement routes a [] _ body =
    { Init = fn r meta _ => genXml r meta,
      Signals = fn _ => return (return {}),
      Edit = fn state => state,
      OnSave = fn _ => return (),
      Validate = fn _ => return (Success {})
    }

con recordRowData = fn t => {
    Default : t,
    Stringify : t -> string,
    Parse : string -> result t,
    Validate : t -> result {}
}

fun parseRecord [fs ::: {Type}] (fl : folder fs)
                (data : $(map recordRowData fs))
                (r : $(map (fn _ => source string) fs))
    : signal (result $fs) =
      @foldR2 [fn _ => source string] [recordRowData] [fn t => signal (result $t)]
       (fn [nm ::_] [t ::_] [rest ::_] [[nm] ~ rest] src datum acc =>
           v' <- dataSignal src datum.Parse
                            (fn x => resultToOpt (datum.Validate x));
           acc' <- acc;
           return (v'' <- v';
                   acc'' <- acc';
                   return ({ nm = v'' } ++ acc'')))
       (return (Success {})) fl r data

(* Turn a list (result x) into an result (list x). Haskell calls this
   'sequence', but that's a keyword. There's no reason this can't be
   polymorphic, but that causes mysterious errors along the lines of
   'polymorphic type needed too late.' *)
fun perform [a ::: Type] (xs : list (result a)) : result (list a) =
    case xs of
        x :: xs' =>
        x' <- x;
        rest <- perform xs';
        return (x' :: rest)
      | [] => return []

fun recordRow [routes ::: {Type}] [meta ::: Type] [fs ::: {Type}]
              (fl : folder fs) (data : $(map recordRowData fs))
    : formElement routes meta fs _ tr = {
    Init =
    let
        (* Build the record of defaults. *)
        val recDefaults : $fs =
            @foldR [recordRowData] [record]
             (fn [nm ::_] [t ::_] [rest ::_] [[nm] ~ rest] v acc =>
                 {nm = v.Default} ++ acc) {} fl data
    in fn _ _ x =>
        (* Stringify each value and create a source containing it. *)
          @foldR2 [recordRowData] [ident]
           [fn t => transaction $(map (fn _ => source string) t)]
           (fn [nm ::_] [t ::_] [rest ::_] [[nm] ~ rest]
                        datum v acc =>
               src <- source (datum.Stringify v);
               acc' <- acc;
               return ({ nm = src } ++ acc'))
           (return {}) fl data (Option.get recDefaults x)
    end,
    Signals = @parseRecord fl data,
    Edit = fn state =>
              @foldR2 [recordRowData] [fn _ => source string] [fn _ => xtr]
               (fn [nm ::_] [t ::_] [rest ::_] [[nm] ~ rest] datum src acc =>
                   <xml>
                     <td>
                     (* The dynClass can't go on the <td> because of a
                           compiler bug(?). *)
                     <div dynClass={
                       colorize (dataSignal src datum.Parse (fn x =>
                                 resultToOpt (datum.Validate x)))}>
                       <ctextbox source={src} class="form-control"/>
                     </div>
                     </td>
                     {acc}
                     </xml>) <xml/> fl data state,
    OnSave = fn _ => return (),
    Validate = fn x => return (
                  @foldR2 [recordRowData] [ident] [fn _ => result {}]
                   (fn [nm ::_] [t ::_] [rest ::_] [[nm] ~ rest] datum v acc =>
                       datum.Validate v; acc)
                  (Success {}) fl data x)


}

fun tableEl [routes ::: {Type}] [nm :: Name] [meta ::: Type]
            [fs ::: {Type}] [state ::: Type]
            (changeRows : bool)
            (element : formElement routes meta fs state tr) = {
      Init = fn r meta xs =>
                states <- Util.listPairMapM
                              (fn m x => (element.Init r m (Some x)))
                              meta.Meta
                              (getOptField [nm] [] xs);
                statesSrc <- source states;
                return {Router = r, Headers = meta.Headers, States = statesSrc,
                        DefaultMeta = meta.Default},
      Signals = fn {States = src, ...} =>
                   states <- signal src;
                   signals <- List.mapM element.Signals states;
                   (* Since each signal can fail individually, we use the result
                      monad to turn a list (signal $fs) into a signal (list
                      $fs). *)
                   return (signals' <- perform signals;
                           return {nm = signals'}),
      Edit = fn state => let
          fun add index =
              old <- get state.States;
              newState <- element.Init state.Router state.DefaultMeta
                                       None;
              set state.States (List.append (List.take (index + 1) old)
                                            (newState :: (List.drop (index + 1)
                                                                    old)))
          fun remove index =
              old <- get state.States;
              set state.States (List.append (List.take index old)
                                       (List.drop (index + 1) old))
          val headerRow = <xml><tr>
            {List.mapX (fn header => <xml><th>{header}</th></xml>)
                       state.Headers}
            (* An extra <td> to match up with the add/remove buttons. *)
            <td/>
          </tr></xml>
          fun buildRow index row = <xml>
            <tr>
              {element.Edit row}
              {if changeRows then <xml><td>
                <div class="btn-group">
                  <button class="btn btn-success"
                  onclick={fn _ => add index} value="+"/>
                  <button class="btn btn-danger"
                  onclick={fn _ => remove index} value="-"/>
                </div>
              </td>
              </xml> else <xml/>}
            </tr>
          </xml>
                in
        <xml>
          {if changeRows then <xml><div>
            <button class="btn btn-success" onclick={fn _ => add (-1)}
              value="Add row"/>
          </div></xml> else <xml/>}
          <dyn signal={states <- signal state.States;
                       return <xml><table class="bs3-table">
                         <thead>{headerRow}</thead>
                         <tbody>{List.mapXi buildRow states}</tbody>
                       </table></xml>}/>
        </xml>
                end,
      OnSave = fn {States = src, ...} =>
                  states <- get src;
                  List.app element.OnSave states,
      Validate = fn xs =>
                    vs <- List.mapM element.Validate xs.nm;
                    return (_ <- perform vs; return {})
}

functor Upload(M : sig
                   end) = struct

    sequence uploadIds

    table uploadTbl : { Id : int, MimeType : string, Content : blob }
                      PRIMARY KEY Id

    fun freshId () = nextval uploadIds

    (* Claim an upload and store it in the database with the given id. *)
    fun saveUpload id handle =
        result <- AjaxUpload.claim handle;
        case result of
            AjaxUpload.NotFound => return ()
          | AjaxUpload.Found found =>
            dml (INSERT INTO uploadTbl (Id, MimeType, Content)
                 VALUES ({[id]}, {[found.MimeType]}, {[found.Content]}))

    fun getUpload (id : int) = Sql.lookup uploadTbl { Id = id }


    val uploadEl = fn [uploadRoute :: Name] [routes ::: {Type}]
                      [[uploadRoute] ~ routes] [a ::: Type]
                      [nm :: Name] mandatory => {
        Init = fn r _ id =>
                  handleSrc <- source None;
                  idSrc <- source (case id of Some id' => id'.nm
                                            | None => None);
                  xml <- AjaxUpload.render {
                         (* None means files auto-upload. *)
                         SubmitLabel = None,
                         OnBegin = return {},
                         OnSuccess = fn handle =>
                                        newId <- rpc (freshId ());
                                        set idSrc (Some newId);
                                        set handleSrc (Some handle),
                         OnError = alert "Upload failed!" };
                  return { IdSrc = idSrc, HandleSrc = handleSrc,
                           UploaderXml = xml },
        Signals = fn state =>
                     id <- signal state.IdSrc;
                     return (if mandatory && Option.isNone id
                             then Error "Must upload a file"
                             else Success { nm = id }),
        Edit = fn state => <xml>
          <div>
            <label class="control-label">Upload</label>
            {state.UploaderXml}
          </div>
        </xml>,
        OnSave = fn state =>
                    id <- get state.IdSrc;
                    handle <- get state.HandleSrc;
                    case handle of
                        (* If there's a handle, we better deal with it. *)
                        Some handle' => (case id of
                            (* Claim handle and associate the data with the
                               given id. *)
                            Some id' => rpc (saveUpload id' handle'); return ()
                          | None => alert "Handle but no ID; shouldn't happen!")
                      (* No handle means there's no data to claim. *)
                      | None => return (),
        Validate = fn id => if mandatory && Option.isNone id.nm then
                                return (Error "Must upload a file")
                            else return (Success {})
        }


end

(* Apply some function to the meta information before using it. *)
fun usingMeta [routes ::: {Type}] [meta ::: Type] [meta' ::: Type]
              [fs ::: {Type}] [state ::: Type] [ctx ::: {Unit}] f formEl = {
    Init = fn r m v => formEl.Init r (f m) v,
    Signals = formEl.Signals,
    Edit = formEl.Edit,
    OnSave = formEl.OnSave,
    Validate = formEl.Validate
    }
