open Bootstrap3
open Template
open Util

con id = int
con cols = [Id = id, Title = string]
con routes = [Create = unit, ViewOne = id, Edit = id, ViewAll = unit]

structure Auth = Authentication
structure BS = Bootstrap3
structure V = Variant

functor Make(M : sig
                 (* The material's data *)
                 con cols :: {Type}
                 val fl : folder cols
                 constraint [Id, Title] ~ cols
                 val validate : $cols -> transaction (Util.result {})
             end) = struct
    sequence ids

    (* Apply [deserialize] to all elements of a record. *)
    fun deserializeRec [fs ::: {Type}] (fl : folder fs)
                       (r : $(map serialized fs)) : $fs =
        @mp [serialized] [ident] (fn [t :::_] => deserialize) fl r

    (* Type of a table row. *)
    con t = cols ++ M.cols
    con t' = [Title = string] ++ M.cols
    con row = map serialized t

    table tbl : row PRIMARY KEY Id

    (* Folder for a row with metadata. *)
    val fl : folder t = @Folder.concat ! M.fl _
    val fl' : folder t' = @Folder.concat ! M.fl _

    (* Serialized material folder. *)
    val safl : folder row = @Folder.mp fl
    (* Witness that serialized materials are injectable. *)
    val sainj : $(map sql_injectable row) =
        @map0 [fn a => sql_injectable (serialized a)]
             (fn [a ::_] => sql_prim) fl

    (* Create a SQL expression corresponding to a row. *)
    fun createSqlExp [env ::: {{Type}}] (data : $t)
        : $(map (sql_exp env [] []) row) =
        @Sql.sqexps safl sainj (
        @mp [ident] [serialized] (fn [t:::_] => serialize) fl data)

    (* Create a new material and insert it into the table. *)
    fun create (rest : $t') =
        v <- M.validate (rest -- #Title);
        case v of
            Success _ =>
            id <- nextval ids;
            let val v = {Id = id} ++ rest in
                dml (insert tbl (createSqlExp v));
                return (Util.Success v)
            end
          | Error err => return (Util.Error err)

    (* Look up an material by its id. *)
    fun lookup (id : id) : transaction (option $t) =
        v <- Sql.lookup tbl {Id = serialize id};
        return (Option.mp (fn x => (@deserializeRec fl
                                     (x ++ {Id = serialize id}))) v)

    fun withMaterial [a] id (f : $t -> transaction a) : transaction a =
        mat <- lookup id;
        case mat of
            Some mat' => f mat'
            | None => error <xml>Not found</xml>

    fun update (id : id) (material : $t) : transaction (result unit) =
        v <- M.validate (material --- cols);
        case v of
            Success _ =>
            delete id;
            dml (insert tbl (createSqlExp material));
            return v
          | _ => return v

    (* Delete the material with a given id. *)
    and delete (id : id) =
        dml (Basis.delete tbl (Sql.selector [#T] {Id = serialize id}))

    val list : transaction (list $t) =
        Monad.mp (List.mp (@deserializeRec fl))
                 (queryL1 (SELECT * FROM tbl ORDER BY tbl.Id DESC))

    functor Gui(R : sig
        val name : string
        con auxRoutes :: {Type}
        val fl : folder auxRoutes
        constraint auxRoutes ~ routes
        val auxHandlers : Linkage.handlers (routes ++ auxRoutes) auxRoutes
        val viewMaterial : Linkage.router (routes ++ auxRoutes) ->
                           option id -> $M.cols -> transaction xbody
        con formState :: Type
        val formEl : Gui.formElement (routes ++ auxRoutes) (option id) M.cols
                                     formState body
      end) = struct

        con routing = variant (routes ++ R.auxRoutes) -> url

        val titleEl = Gui.textboxEl [#Title] "Title" "New title" nonEmpty

        val formEl = Gui.concatForm titleEl R.formEl

        (* View a single material. *)
        fun viewOne (r : routing) (id : id) : transaction page =
            data <- lookup id;
            case data of
                None => error <xml>Not found</xml>
              | Some data' =>
                rendered <- R.viewMaterial r (Some id) (data' -- #Id -- #Title);
                template {
                Inner = rendered,
                Crumbs = {Url = Some (r (make [#ViewAll] ())),
                          Title = R.name} ::
                         {Url = None, Title = data'.Title} :: []}
        and editor (r : routing) (mat : option $t) : transaction page =
            let
                val v : option $t' = Option.mp (fn x => x -- #Id) mat
                val maybeId : option id = Option.mp (fn x => x.Id) mat
                val itemCrumb = case mat of
                    Some mat' => {Url = Some (r (make [#ViewOne] mat'.Id)),
                                  Title = mat'.Title}
                  | None => {Url = None, Title = "New item"}
            in
            Auth.requireTeacher;
            state <- formEl.Init r maybeId v;
            titleSrc <- source (getOptField [#Title] "New item" v);
            titleModified <- source False;
            (* Hold the ID, if any. We need this so that when we save an
               existing item, further 'save' clicks don't keep creating new
               items. *)
            idSrc <- source maybeId;
            savingSrc <- source False;
            saveButton <- Widget.formElSaveButton formEl state
                     (fn v =>
                         currentId <- get idSrc;
                         case currentId of
                             Some id =>
                             _ <- rpc (update id (v ++ { Id = id }));
                             return ()
                           | None =>
                             new <- rpc (create v);
                             case new of
                                 Util.Success new' =>
                                 set idSrc (Some new'.Id);
                                 return ()
                               | Util.Error _ =>
                                 return ());
            template {
            Inner = <xml>
              {formEl.Edit state}
              <div>{saveButton}</div>
            </xml>,
            Crumbs = {Url = Some (r (make [#ViewAll] ())), Title = R.name} ::
                     itemCrumb ::
                     {Url = None, Title = "Edit"} :: []}
            end


        and edit (r : routing) (id : id) : transaction page =
            Auth.requireTeacher;
            withMaterial id (fn mat => editor r (Some mat))

        and deleteRpc (id : id) : transaction unit =
            Auth.requireTeacher; delete id

        and makeNew (r : routing) () : transaction page =
            Auth.requireTeacher;
            editor r None

        and showAll (r : routing) () : transaction page =
            items <- list;
            createLink <- Authentication.ifTeacher <xml>
              <a class={classes BS.btn BS.btn_primary}
              href={r (make [#Create] ())}>
              Create
              </a>
            </xml>;
            inner <- List.mapXM (fn item =>
                editLink <- Auth.ifTeacher <xml>
                  <a class="btn btn-default" href={r (make [#Edit] item.Id)}>
                    Edit
                  </a>
                </xml>;
                deleteLink <- Auth.ifTeacher <xml>
                  <button class="btn btn-danger" onclick={fn _ =>
                      reallyDelete <-
                      confirm ("Are you sure you want to delete \"" ^
                               item.Title ^ "\"?");
                      if reallyDelete then
                          rpc (deleteRpc item.Id);
                          redirect (r (make [#ViewAll] ()))
                      else return ()}
                    value="Delete"/>
                </xml>;
                return <xml>
                  <h2><a href={r (make [#ViewOne] item.Id)}>
                    {[item.Title]}
                    {editLink}
                    {deleteLink}
                  </a></h2>
                </xml>) items;
            template {
            Inner = join inner createLink,
            Crumbs = {Url = None, Title = R.name} :: []}

        val routingRecord : Linkage.handlers (routes ++ R.auxRoutes) routes = {
             Create = makeNew, ViewOne = viewOne, Edit = edit, ViewAll = showAll
        }

        val handlers : Linkage.closable (routes ++ R.auxRoutes) =
            routingRecord ++ R.auxHandlers

        structure Closed = Linkage.Close(struct
                         con provides = routes ++ R.auxRoutes
                         val fl = @Folder.concat ! _ R.fl
                         val fs = handlers
                     end)

        val routed = Closed.routed
    end
end
