(* An material is anything that teachers can create and edit and students can
   view. (Better names accepted.) Examples include file uploads and lecture
   notes. Things that are graded should use Gradable instead. *)

con id
con cols = [Id = id, Title = string]
con routes = [Create = unit, ViewOne = id, Edit = id, ViewAll = unit]


(* Make an material module. This module returns functions for working with the
   materials, and then a second functor that generates the GUI. The module is
   separated in two so that the template can use data functions to, e.g., have a
   dropdown list of all materials. *)
functor Make(M : sig
                 (* The material's data. *)
                 con cols :: {Type}
                 val fl : folder cols
                 constraint [Id, Title] ~ cols
                 (* A validator, which should return None on success and (Some
                    err) on an error. *)
                 val validate : $cols -> transaction (Util.result {})
             end) : sig
    (* The type of an material with metadata. *)
    con t = [Id = id, Title = string] ++ M.cols
    (* The type of an material with metadata, but without ID. Useful when
       creating a new material. *)
    con t' = [Title = string] ++ M.cols
    val fl : folder t

    (* The table of materials, with the ID as a primary key. *)
    val tbl : Basis.sql_table (map serialized t) [Pkey = [Id]]

    (* Create a new material. Returns the new material or a validation failure
       reason. *)
    val create : $t' -> transaction (Util.result $t)
    (* Look up an material by ID. *)
    val lookup : id -> transaction (option $t)
    (* Perform some action with the material with the given ID; errors if no
       such material exists. *)
    val withMaterial : a ::: Type -> id -> ($t -> transaction a) ->
                       transaction a

    (* Update the data for an material. If validation on the new data fails, the
       old data is not deleted. *)
    val update : id -> $t -> transaction (Util.result unit)
    (* Delete an material; does nothing if there is no material with that id. *)
    val delete : id -> transaction unit

    (* List all materials. *)
    val list : transaction (list $t)

    functor Gui(R : sig
        (* The name of this kind of material. *)
        val name : string
        con auxRoutes :: {Type}
        val fl : folder auxRoutes
        constraint auxRoutes ~ routes
        val auxHandlers : Linkage.handlers (routes ++ auxRoutes) auxRoutes
        val viewMaterial : Linkage.router (routes ++ auxRoutes) ->
                           option id -> $M.cols -> transaction xbody
        con formState :: Type
        val formEl : Gui.formElement (routes ++ auxRoutes) (option id) M.cols
                                     formState body
      end) : sig
        val routed : $(map (fn t => t -> transaction page)
                               (routes ++ R.auxRoutes))
    end
end
